# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Physical metrics for tensor maps and images."""

from __future__ import annotations
from skais_mapper.profile import (
    _sanitize_ndim,
    _make_grid,
    compute_centers,
    half_mass_radius,
    radial_pdf,
    radial_cdf,
)
from skais_mapper._compat import TORCH_AVAILABLE
from typing import Literal, Callable, TYPE_CHECKING

if TORCH_AVAILABLE or TYPE_CHECKING:
    import torch
    import torch.nn.functional as F
else:
    from skais_mapper import _torch_stub as __stub  # noqa
    from skais_mapper._torch_stub import *  # noqa: F401,F403


__all__ = [
    "CenterOffsetError",
    "RadialProfileCurveError",
    "MapTotalError",
    "AsymmetryError",
    "ClumpinessError",
    "PowerSpectrumError",
]


def _aperture_mask(
    B: int,
    H: int,
    W: int,
    centers: torch.Tensor,
    r_in: torch.Tensor,
    r_out: torch.Tensor,
    device: torch.device | str | None = None,
    dtype: torch.dtype | None = None,
) -> torch.Tensor:
    """Annular mask between r_in and r_out (inclusive) per sample of shape (B, H, W).

    Args:
        B: Batch size.
        H: Mask height.
        W: Mask width.
        centers: Center coordinates of shape (B, 2) or (B, H, W).
        r_in: Inner radius of shape (B,).
        r_out: Outer radius of shape (B,).
        device: Tensor allocation/computation device.
        dtype: Tensor data type.
    """
    device = torch.device(device) if device is not None else torch.device("cpu")
    dtype = dtype if dtype is not None else torch.float32
    XX, YY = _make_grid(W, H, device=device, dtype=dtype)
    dx = XX - centers[:, 0].view(B, 1, 1)
    dy = YY - centers[:, 1].view(B, 1, 1)
    r = torch.sqrt(dx * dx + dy * dy)
    mask = (r >= r_in.view(B, 1, 1)) & (r <= r_out.view(B, 1, 1))
    return mask.to(dtype)


class CenterOffsetError:
    """Center offset between two maps: center-of-mass or peak position distance.

    Distances can be reported in pixels (scaled by pixel_size) or normalized by:
      - image_radius: half the image diagonal
      - half-mass radius: with respect to a reference map
    """

    def __init__(
        self,
        center: Literal["com", "peak"] = "com",
        normalize: Literal["image_radius", "r50"] = "image_radius",
        pixel_size: float = 1.0,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
        reduction: Callable | None = torch.mean,
    ):
        """Constructor.

        Args:
            center: Centering mode; one of ["com", "peak"].
            normalize: Normalization mode; one of ["image_radius", "r50"].
            pixel_size: Physical size per pixel (multiplies distances).
            eps: Numerical stability for divisions.
            n_observations: Number of observations seen by the internal state.
            device: Tensor allocation/computation device.
            reduction: Reduction function to be used when computing metric scalar.
        """
        self.reduction = reduction
        self.device = torch.get_default_device() if device is None else device
        if center not in ("com", "peak"):
            raise ValueError("Input center must be 'com' or 'peak'")
        if normalize not in ("image_radius", "r50"):
            raise ValueError("Input normalize must be 'image_radius' or 'r50'")
        self.center = center
        self.normalize = normalize
        self.pixel_size = float(pixel_size)
        self.eps = float(eps)
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(1, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(1, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(1, device=self.device)

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = self.aggregate.to(device=self.device)
        self.max_aggregate = self.max_aggregate.to(device=self.device)
        self.min_aggregate = self.min_aggregate.to(device=self.device)

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(self.nbins, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(self.nbins, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(self.nbins, device=self.device)

    @staticmethod
    def _com_xy(
        maps: torch.Tensor,
        X: torch.Tensor | None = None,
        Y: torch.Tensor | None = None,
        eps: float = 1e-12,
    ) -> torch.Tensor:
        """Compute center-of-mass coordinates for a batch of images.

        Args:
            maps: Input maps of shape (B, H, W), (B, C, H, W), or (H, W).
            X: X coordinate grid of shape (H, W).
            Y: Y coordinate grid of shape (H, W).
            eps: Numerical stability for divisions.
        """
        maps_ = _sanitize_ndim(maps)
        if X is None or Y is None:
            B, H, W = maps_.shape
            X, Y = _make_grid(W, H, device=maps.device, dtype=maps.dtype)
        return compute_centers(maps_, X, Y, mode="centroid", eps=eps)

    @staticmethod
    def _peak_xy(maps: torch.Tensor) -> torch.Tensor:
        """Compute peak coordinates for a batch of images.

        Args:
            maps: Input maps of shape (B, H, W), (B, C, H, W), or (H, W).
        """
        maps_ = _sanitize_ndim(maps)
        B, H, W = maps_.shape
        peaks = torch.argmax(maps_, dim=(-2, -1))
        return peaks

    @staticmethod
    def _half_mass_radius(maps: torch.Tensor, eps: float = 1e-12) -> torch.Tensor:
        """Compute half-mass radius for a batch of images.

        Args:
            maps: Input maps of shape (B, H, W), (B, C, H, W), or (H, W).
            eps: Numerical stability for divisions.
        """
        return half_mass_radius(maps, eps=eps)

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Aggregate batch center offsets.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if targ_.shape != pred_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}.")
        B, H, W = pred_.shape
        if self.center == "com":
            centers_t = self._com_xy(targ_)
            centers_p = self._com_xy(pred_)
        else:
            centers_t = self._peak_xy(targ_)
            centers_p = self._peak_xy(pred_)
        delta2 = (centers_p - centers_t) ** 2
        dist = torch.sqrt(delta2.sum(dim=1)) * self.pixel_size
        if self.normalize == "image_radius":
            denom = (
                0.5
                * torch.sqrt(torch.tensor(H**2 + W**2, dtype=dist.dtype, device=dist.device))
                * self.pixel_size
            )
        elif self.normalize == "r50":
            denom = self._half_mass_radius(targ_) * self.pixel_size
        else:
            denom = 1
        dist = dist / denom.clamp_min(self.eps)
        self.aggregate += dist.sum(dim=0)
        self.min_aggregate = torch.min(self.min_aggregate, dist.amin())
        self.max_aggregate = torch.max(self.max_aggregate, dist.amax())
        self.n_observations += B

    @torch.no_grad()
    def compute(self, reduction: Callable | None = None) -> torch.Tensor:
        """Return the center offset error over all seen samples."""
        if self.n_observations == 0:
            return self.aggregate
        if reduction is None:
            reduction = self.reduction
        return reduction(self.aggregate / float(self.n_observations))


class RadialProfileCurveError:
    """Shape error between radial density profiles of predicted and target maps.

    This metric compares the azimuthally-averaged radial distribution shapes by
    computing a radial probability density function (PDF) for each input map and
    integrating the difference across radius.

    Note: This metric is insensitive to global normalization/flux differences.
    Use `MapTotalError` to capture amplitude discrepancies.
    """

    def __init__(
        self,
        nbins: int = 100,
        center_mode: Literal["centroid", "image_center", "fixed"] = "image_center",
        cumulative: bool = False,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
        reduction: Callable | None = torch.sum,
    ) -> None:
        """Constructor:

        Args:
            nbins: Number of radial bins.
            center_mode: Centering mode for radial profiles; one of
              `["centroid", "image_center", "fixed"]`.
            cumulative: Whether to compare cumulative radial profiles.
            eps: Numerical stability for divisions.
            n_observations: Number of observations (bins) seen by the internal state.
            device: Tensor allocation/computation device.
            reduction: Reduction function to be used when computing metric scalar.
        """
        self.reduction = reduction
        self.device = torch.get_default_device() if device is None else device
        self.nbins = int(max(1, nbins))
        self.center_mode = center_mode
        self.cumulative = bool(cumulative)
        self.eps = float(eps)
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(self.nbins, device=self.device)
        self.lsq_aggregate = torch.zeros(self.nbins, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(nbins, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(nbins, device=self.device)

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = self.aggregate.to(device=self.device)
        self.lsq_aggregate = self.lsq_aggregate.to(device=self.device)
        self.max_aggregate = self.max_aggregate.to(device=self.device)
        self.min_aggregate = self.min_aggregate.to(device=self.device)

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(self.nbins, device=self.device)
        self.lsq_aggregate = torch.zeros(self.nbins, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(self.nbins, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(self.nbins, device=self.device)

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Accumulate batch errors.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if targ_.shape != pred_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}")
        B, H, W = pred_.shape
        # Compute radial PDFs for both maps using identical binning.
        if self.cumulative:
            pdf_p, edges = radial_cdf(pred_, nbins=self.nbins, center_mode=self.center_mode)
            pdf_t, _ = radial_cdf(targ_, bin_edges=edges[0], center_mode=self.center_mode)
        else:
            pdf_p, edges = radial_pdf(pred_, nbins=self.nbins, center_mode=self.center_mode)
            pdf_t, _ = radial_pdf(targ_, bin_edges=edges[0], center_mode=self.center_mode)
        dr = (edges[:, 1:] - edges[:, :-1]).clamp_min(self.eps)
        diff = (pdf_p - pdf_t).abs()
        per_bin_err = diff * dr
        self.n_observations += B
        self.aggregate += torch.sum(per_bin_err, dim=0)
        self.lsq_aggregate += torch.sum(per_bin_err.pow(2), dim=0)
        self.min_aggregate = torch.min(self.min_aggregate, per_bin_err.amin(dim=0))
        self.max_aggregate = torch.max(self.max_aggregate, per_bin_err.amax(dim=0))

    @property
    def mean_per_bin(self) -> torch.Tensor:
        """Mean error per bin."""
        return self.aggregate / self.n_observations

    @property
    def var_per_bin(self) -> torch.Tensor:
        """Error variance per bin."""
        return (self.lsq_aggregate / self.n_observations) - self.mean_per_bin.pow(2)

    @property
    def std_per_bin(self) -> torch.Tensor:
        """Error variance per bin."""
        return self.var_per_bin.sqrt() / self.n_observations

    @torch.no_grad()
    def compute(self, reduction: Callable | None = None) -> torch.Tensor:
        """Return the radial profile curve error reduced to a scalar."""
        if reduction is None:
            reduction = self.reduction
        return reduction(self.mean_per_bin)


class MapTotalError:
    """Absolute or relative error in total integrated map quantity (e.g., flux/mass)."""

    def __init__(
        self,
        relative: bool = True,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
    ) -> None:
        """Constructor.

        Args:
            relative: If True, returns absolute fractional error:
              |sum(pred) - sum(target)| / (|sum(target)| + eps).
              If False, returns absolute error |sum(pred) - sum(target)|.
            eps: Numerical stability for relative error denominator.
            n_observations: Number of observations seen by the internal state.
            device: Tensor allocation/computation device.
        """
        self.relative = bool(relative)
        self.eps = float(eps)
        self.device = torch.get_default_device() if device is None else device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(1, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(1, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(1, device=self.device)

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = self.aggregate.to(device=self.device)
        self.max_aggregate = self.max_aggregate.to(device=self.device)
        self.min_aggregate = self.min_aggregate.to(device=self.device)

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(1, device=self.device)
        self.max_aggregate = -torch.inf * torch.ones(1, device=self.device)
        self.min_aggregate = torch.inf * torch.ones(1, device=self.device)

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Accumulate batch errors.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if pred_.shape != targ_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}.")
        B, H, W = pred_.shape
        sum_p = pred_.flatten(1).sum(dim=1)
        sum_t = targ_.flatten(1).sum(dim=1)
        if self.relative:
            per_sample_err = (sum_p - sum_t).abs() / (sum_t.abs() + self.eps)
        else:
            per_sample_err = (sum_p - sum_t).abs()
        self.aggregate += per_sample_err.sum(dim=0)
        self.min_aggregate = torch.min(self.min_aggregate, per_sample_err.amin())
        self.max_aggregate = torch.max(self.max_aggregate, per_sample_err.amax())
        self.n_observations += B

    def compute(self) -> torch.Tensor:
        """Return the mean total quantity error over all seen samples."""
        if self.n_observations == 0:
            return self.aggregate
        return self.aggregate / float(self.n_observations)


class AsymmetryError:
    """Asymmetry (A) metric difference between predicted and target maps.

    A = sum(|I - I_180| within aperture) / (sum(|I| within aperture) + eps)

    The 180-degree rotation is performed about a chosen center:
    - center_mode="image_center": uses geometric image center and fast flip.
    - center_mode="centroid": uses center-of-mass per map; rotation via bilinear sampling.

    Aperture:
    - aperture="full": use the entire image.
    - aperture="r_factor": use a circular aperture with radius r_factor * R50,
      where R50 is the half-mass radius measured on the same map and center.
    """

    def __init__(
        self,
        center_mode: Literal["image_center", "centroid"] = "image_center",
        aperture: Literal["full", "r_factor"] = "r_factor",
        r_factor: float = 1.5,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
        reduction: Callable | None = torch.mean,
    ) -> None:
        """Constructor.

        Args:
            center_mode: Centering mode for rotation; one of `["centroid", "image_center"]`.
            aperture: Aperture for error calculation; one of `["full" or "r_factor"]`.
            r_factor: Multiplier for R50 if `aperture="r_factor"`.
            eps: Numerical stability.
            n_observations: Number of observations seen by the internal state.
            device: Tensor allocation/computation device.
            reduction: Reduction function to be used when computing metric scalar.
        """
        if center_mode not in ("image_center", "centroid"):
            raise ValueError("center_mode must be 'image_center' or 'centroid'")
        if aperture not in ("full", "r_factor"):
            raise ValueError("aperture must be 'full' or 'r_factor'")
        self.device = torch.get_default_device() if device is None else device
        self.reduction = reduction
        self.center_mode = center_mode
        self.aperture = aperture
        self.r_factor = float(r_factor)
        self.eps = float(eps)
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = None
        self.target_aggregate = None

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = (
            self.aggregate.to(device=self.device) if self.aggregate is not None else None
        )
        self.target_aggregate = (
            self.target_aggregate.to(device=self.device)
            if self.target_aggregate is not None
            else None
        )

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = None
        self.target_aggregate = None

    @staticmethod
    def _rotate_about_center(
        maps: torch.Tensor,
        center: torch.Tensor,
        eps: float = 1e-12,
    ) -> torch.Tensor:
        """Rotate maps by 180 degrees about a given center.

        Args:
            maps: Input maps of shape (B, H, W), (B, C, H, W), or (H, W).
            center: Center coordinates of shape (B, 2).
            eps: Numerical stability for divisions.
        """
        maps_ = _sanitize_ndim(maps)
        B, H, W = maps_.shape
        XX, YY = _make_grid(W, H, device=maps_.device, dtype=maps_.dtype)
        x_prime = 2 * center[:, 0].view(B, 1, 1) - XX
        y_prime = 2 * center[:, 1].view(B, 1, 1) - YY
        x_norm = 2 * x_prime / max(W - 1, 1) - 1
        y_norm = 2 * y_prime / max(H - 1, 1) - 1
        grid = torch.stack((x_norm, y_norm), dim=-1)
        maps_ = maps_.unsqueeze(1)
        rot = F.grid_sample(maps_, grid, mode="bilinear", padding_mode="border", align_corners=True)
        return rot[:, 0, :, :]

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Accumulate batch asymmetry errors.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if pred_.shape != targ_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}.")
        B, H, W = pred_.shape
        # Centers
        XX, YY = _make_grid(W, H, device=targ_.device, dtype=targ_.dtype)
        centers_t = compute_centers(targ_, XX, YY, eps=self.eps, mode=self.center_mode)
        centers_p = compute_centers(pred_, XX, YY, eps=self.eps, mode=self.center_mode)
        # Rotate maps by 180 degrees about their centers
        if self.center_mode == "image_center":
            targ_rot = torch.flip(targ_, dims=(-2, -1))
            pred_rot = torch.flip(pred_, dims=(-2, -1))
        elif self.center_mode == "centroid":
            targ_rot = self._rotate_about_center(targ_, centers_t, eps=self.eps)
            pred_rot = self._rotate_about_center(pred_, centers_p, eps=self.eps)
        # Aperture masks
        if self.aperture == "full":
            mask_p = torch.ones_like(pred_)
            mask_t = torch.ones_like(targ_)
        elif self.aperture == "r_factor":
            # r_factor * R50 using each map's own center
            r50_p = half_mass_radius(pred_, eps=self.eps, center_mode=self.center_mode)
            r50_t = half_mass_radius(targ_, eps=self.eps, center_mode=self.center_mode)
            r_out_p = (self.r_factor * r50_p).clamp_min(1.0)
            r_out_t = (self.r_factor * r50_t).clamp_min(1.0)
            zero = torch.zeros_like(r_out_p)
            mask_p = _aperture_mask(
                B, H, W, centers_p, zero, r_out_p, device=targ_.device, dtype=targ_.dtype
            )
            mask_t = _aperture_mask(
                B, H, W, centers_t, zero, r_out_t, device=targ_.device, dtype=targ_.dtype
            )
        # Asymmetry maps
        err_p = mask_p * (pred_ - pred_rot).abs()
        rel_p = (mask_p * pred_.abs()).flatten(1).sum(dim=1)
        A_p = err_p / rel_p.clamp_min(self.eps).view(B, 1, 1)
        err_t = mask_t * (targ_ - targ_rot).abs()
        rel_t = (mask_t * targ_.abs()).flatten(1).sum(dim=1)
        A_t = err_t / rel_t.clamp_min(self.eps).view(B, 1, 1)
        # Aggregate per sample asymmetries
        if self.aggregate is None:
            self.aggregate = A_p.sum(dim=0)
        else:
            self.aggregate += A_p.sum(dim=0)
        if self.target_aggregate is None:
            self.target_aggregate = A_t.sum(dim=0)
        else:
            self.target_aggregate += A_t.sum(dim=0)
        self.n_observations += B

    def compute(self, reduction: Callable | None = None) -> torch.Tensor | None:
        """Return the mean total quantity error over all seen samples."""
        if reduction is None:
            reduction = self.reduction
        if self.n_observations == 0:
            return self.aggregate
        diff = (self.aggregate - self.target_aggregate).abs()
        return reduction(diff.flatten() / float(self.n_observations))


class ClumpinessError:
    """Smoothness/Clumpiness difference between two different 2D maps.

    S = sum(max(I - I_s, 0) within annulus) / (sum(I within annulus) + eps)

    The smoothed map I_s is computed via Gaussian blur with a configurable sigma:
    - sigma_mode="pixels": sigma is given in pixels (constant for all samples).
    - sigma_mode="r_factor": sigma = s_factor * R50 (per sample), where R50 is
      measured around the chosen center (image center or centroid).

    The annulus is defined by [r_inner, r_outer]:
    - r_inner = inner_excision * R50  (commonly ~0.2)
    - r_outer = aperture_factor * R50 (commonly ~1.5)
      If aperture_factor <= inner_excision, the annulus collapses to empty.
    """

    def __init__(
        self,
        center_mode: Literal["image_center", "centroid"] = "image_center",
        sigma_mode: Literal["pixels", "r_factor"] = "pixels",
        sigma_pixels: float = 1.0,
        r_factor: float = 1.5,
        inner_excision: float = 0.2,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
        reduction: Callable | None = torch.mean,
    ) -> None:
        """Constructor.

        Args:
            center_mode: Centering mode for defining R50 and annulus center;
              one of `["centroid", "image_center"]`..
            sigma_mode: "pixels" or "r_factor" for the smoothing scale.
            sigma_pixels: Gaussian sigma in pixels when sigma_mode="pixels".
            r_factor: Outer radius multiple of R50 for the annulus (e.g., 1.5).
            inner_excision: Inner exclusion multiple of R50 (e.g., 0.2).
            eps: Numerical stability.
            n_observations: Number of observations seen by the internal state.
            device: Tensor allocation/computation device.
            reduction: Reduction function to be used when computing metric scalar.
        """
        if center_mode not in ("image_center", "centroid"):
            raise ValueError("center_mode must be 'image_center' or 'centroid'")
        if sigma_mode not in ("pixels", "r_factor"):
            raise ValueError("sigma_mode must be 'pixels' or 'r_factor'")
        self.device = torch.get_default_device() if device is None else device
        self.reduction = reduction
        self.center_mode = center_mode
        self.sigma_mode = sigma_mode
        self.sigma_pixels = float(sigma_pixels)
        self.r_factor = float(r_factor)
        self.inner_excision = float(inner_excision)
        self.eps = float(eps)
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = None
        self.target_aggregate = None

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = (
            self.aggregate.to(device=self.device) if self.aggregate is not None else None
        )
        self.target_aggregate = (
            self.target_aggregate.to(device=self.device)
            if self.target_aggregate is not None
            else None
        )

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = None
        self.target_aggregate = None

    def sigma(
        self,
        batch_size: int,
        r50: torch.Tensor | None = None,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        """Get the Gaussian kernel sigma tensor for blurring."""
        device = torch.device(device) if device is not None else torch.device("cpu")
        dtype = dtype if dtype is not None else torch.float32
        if self.sigma_mode == "pixels":
            return torch.full((batch_size,), self.sigma_pixels, device=device, dtype=dtype)
        elif self.sigma_mode == "r_factor":
            if r50 is None:
                raise ValueError("r50 must be provided when sigma_mode='r_factor'.")
            return (self.r_factor * r50).clamp_min(0.1).to(device=device, dtype=dtype)

    @staticmethod
    def _gaussian_kernel2d(
        sigma: float,
        device: torch.device | str | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        """Create a 2D Gaussian kernel normalized to sum=1 of size = 2*ceil(3*sigma)+1."""
        device = torch.device(device) if device is not None else torch.device("cpu")
        dtype = dtype if dtype is not None else torch.float32
        if sigma <= 0:
            # Degenerate kernel -> identity in convolution; we'll skip conv in that case.
            k = torch.zeros((1, 1), dtype=dtype, device=device)
            k[0, 0] = 1.0
            return k
        radius = int(torch.ceil(torch.tensor(3.0 * sigma)).item())
        ax = torch.arange(-radius, radius + 1, device=device, dtype=dtype)
        xx, yy = torch.meshgrid(ax, ax, indexing="ij")
        kernel = torch.exp(-(xx * xx + yy * yy) / (2.0 * sigma * sigma))
        kernel = kernel / kernel.sum()
        return kernel

    @staticmethod
    def _gaussian_blur(
        maps: torch.Tensor,
        sigma: torch.Tensor,
        eps: float = 1e-12,
    ) -> torch.Tensor:
        """Rotate maps by 180 degrees about a given center.

        Args:
            maps: Input maps of shape (B, H, W), (B, C, H, W), or (H, W).
            sigma: Gaussian kernel sigmas (B,).
            eps: Numerical stability for divisions.
        """
        maps_ = _sanitize_ndim(maps)
        B, H, W = maps_.shape
        blurred = []
        for i in range(B):
            if sigma[i] <= 0:
                blurred.append(maps_[i : i + 1, :, :])
                continue
            k = ClumpinessError._gaussian_kernel2d(
                sigma=sigma[i].item(), dtype=maps_.dtype, device=maps_.device
            )
            k = k.view(1, 1, k.shape[0], k.shape[1])
            pad_h = (k.shape[2] - 1) // 2
            pad_w = (k.shape[3] - 1) // 2
            x = F.pad(maps[i].unsqueeze(0), (pad_w, pad_w, pad_h, pad_h), mode="reflect")
            y = F.conv2d(x, k)
            blurred.append(y.squeeze(1))
        return torch.cat(blurred, dim=0)

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Accumulate batch asymmetry errors.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if pred_.shape != targ_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}.")
        B, H, W = targ_.shape
        # Aperture mask
        XX, YY = _make_grid(W, H, device=targ_.device, dtype=targ_.dtype)
        centers = compute_centers(targ_, XX, YY, eps=self.eps, mode=self.center_mode)
        r50_t = half_mass_radius(targ_, eps=self.eps, center_mode=self.center_mode)
        r_in = (self.inner_excision * r50_t).clamp_min(0.0)
        r_out = (self.r_factor * r50_t).clamp_min(1.0)
        ann_mask = _aperture_mask(
            B, H, W, centers, r_in, r_out, device=targ_.device, dtype=targ_.dtype
        )
        # Smoothing sigmas
        sigmas_t = self.sigma(B, r50=r50_t, device=targ_.device, dtype=targ_.dtype)
        sigmas_p = sigmas_t
        # Blur each sample (per-sample sigma)
        blur_t = self._gaussian_blur(targ_, sigmas_t, self.eps)
        blur_p = self._gaussian_blur(pred_, sigmas_p, self.eps)
        # Smoothness values per sample (positive residuals only)
        resid_p = ann_mask * (pred_ - blur_p).clamp_min(0.0)
        rel_p = (ann_mask * pred_.abs()).flatten(1).sum(dim=1)
        resid_t = ann_mask * (targ_ - blur_t).clamp_min(0.0)
        rel_t = (ann_mask * targ_.abs()).flatten(1).sum(dim=1)
        S_p = resid_p / rel_p.clamp_min(self.eps).view(B, 1, 1)
        S_t = resid_t / rel_t.clamp_min(self.eps).view(B, 1, 1)
        # Aggregate per sample smoothness
        if self.aggregate is None:
            self.aggregate = S_p.sum(dim=0)
        else:
            self.aggregate += S_p.sum(dim=0)
        if self.target_aggregate is None:
            self.target_aggregate = S_t.sum(dim=0)
        else:
            self.target_aggregate += S_t.sum(dim=0)
        self.n_observations += B

    def compute(self, reduction: Callable | None = None) -> torch.Tensor | None:
        """Return the mean total quantity error over all seen samples."""
        if reduction is None:
            reduction = self.reduction
        if self.n_observations == 0:
            return self.aggregate
        diff = (self.aggregate - self.target_aggregate).abs()
        return reduction(diff.flatten() / float(self.n_observations))


class PowerSpectrumError:
    """Error between radially averaged 2D power spectra.

    This metric can optionally detrend, apply windows (to avoid edge leakage),
    and computes the power spectrum of 2D maps via FFT with orthogonal norm.
    The power spectra are radially averaged to obtain P(k) curves between kmin and kmax.
    """

    def __init__(
        self,
        nbins: int = 64,
        kmin: float | None = None,
        kmax: float | None = None,
        log_bins: bool = False,
        k_mode: Literal["cycles", "angular"] = "cycles",
        pixel_size: float = 1.0,
        detrend: bool = True,
        window: Literal["hann"] | None = "hann",
        log_power: bool = False,
        per_bin_weighted: bool = True,
        freeze_edges: bool = True,
        eps: float = 1e-12,
        n_observations: int = 0,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        reduction: Callable | None = torch.mean,
    ) -> None:
        """Constructor.

        Args:
            nbins: Number of radial k-bins.
            kmin: Minimum k (inclusive). If `None`, uses smallest positive k on the grid.
            kmax: Maximum k (exclusive upper edge). If `None`, uses Nyquist.
            log_bins: If `True`, use logarithmically spaced bin edges in `[kmin, kmax]`.
              `kmin` must be > 0 for log bins; will be clamped to min positive k.
            k_mode: "cycles" for cycles per unit length (default), or
              "angular" for 2π times cycles.
            pixel_size: Physical size per pixel. Units set k-units.
            detrend: Subtract per-image mean before FFT.
            window: "hann" or None.
            log_power: If `True`, compare log10(P + eps) curves instead of linear power.
            per_bin_weighted: If `True`, integrate curve error using bin width dk.
            freeze_edges: If `True`, bin edges and k-bin assignment are fixed
              after first batch.
            eps: Numerical stability constant.
            n_observations: Number of observations seen by the internal state.
            device: Tensor allocation/computation device.
            dtype: Tensor data type for internal tensors.
            reduction: Reduction function to be used when computing metric scalar.
        """
        if window not in ("hann", None):
            raise ValueError("window must be 'hann' or None")
        if k_mode not in ("cycles", "angular"):
            raise ValueError("k_mode must be 'cycles' or 'angular'")
        self.nbins = int(max(1, nbins))
        self.kmin = kmin
        self.kmax = kmax
        self.log_bins = bool(log_bins)
        self.k_mode = k_mode
        self.pixel_size = float(pixel_size)
        self.detrend = bool(detrend)
        self.window = window
        self.log_power = bool(log_power)
        self.per_bin_weighted = bool(per_bin_weighted)
        self.freeze_edges = bool(freeze_edges)
        self.device = torch.get_default_device() if device is None else device
        self.dtype = dtype if dtype is not None else torch.float32
        self.reduction = reduction
        self.eps = float(eps)
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(self.nbins, device=self.device, dtype=self.dtype)
        self.lsq_aggregate = torch.zeros(self.nbins, device=self.device, dtype=self.dtype)
        self.max_aggregate = -torch.inf * torch.ones(
            self.nbins, device=self.device, dtype=self.dtype
        )
        self.min_aggregate = torch.inf * torch.ones(
            self.nbins, device=self.device, dtype=self.dtype
        )

        self._W: int = None  # Width of the last seen image.
        self._H: int = None  # Height of the last seen image.
        self._k_edges: torch.Tensor = None  # Radial k-bin edges.
        self._dk: torch.Tensor = None  # Width of each k-bin.
        self._bin_index: torch.Tensor = None  # Pixel-to-bin assignment.
        self._hann2d: torch.Tensor = None  # Hann window for the last seen image.

    def to(self, device: torch.device):
        """Perform tensor device conversion for all internal tensors.

        Args:
            device: Tensor allocation/computation device.
        """
        self.device = device
        self.n_observations = self.n_observations.to(device=self.device)
        self.aggregate = self.aggregate.to(device=self.device)
        self.lsq_aggregate = self.lsq_aggregate.to(device=self.device)
        self.max_aggregate = self.max_aggregate.to(device=self.device)
        self.min_aggregate = self.min_aggregate.to(device=self.device)

    def reset(self, n_observations: int = 0, device: torch.device | None = None) -> None:
        """Reset internal metrics state."""
        self.device = device if device is not None else self.device
        self.n_observations = torch.tensor(n_observations, device=self.device)
        self.aggregate = torch.zeros(self.nbins, device=self.device, dtype=self.dtype)
        self.lsq_aggregate = torch.zeros(self.nbins, device=self.device, dtype=self.dtype)
        self.max_aggregate = -torch.inf * torch.ones(
            self.nbins, device=self.device, dtype=self.dtype
        )
        self.min_aggregate = torch.inf * torch.ones(
            self.nbins, device=self.device, dtype=self.dtype
        )
        self._W = None
        self._H = None
        self._k_edges = None
        self._dk = None
        self._bin_index = None
        self._hann2d = None

    @staticmethod
    def _hann_window_1d(
        N: int,
        device: torch.device | str | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        """Create a 1D Hann window."""
        device = torch.device(device) if device is not None else torch.device("cpu")
        dtype = dtype = dtype if dtype is not None else torch.float32
        if N <= 1:
            return torch.ones((N,), dtype=dtype, device=device)
        try:
            return torch.hann_window(N, periodic=False, dtype=dtype, device=device)
        except AttributeError:  # Fallback for older PyTorch versions
            n = torch.arange(0, N, dtype=dtype, device=device)
            return 0.5 * (1 - torch.cos(2 * torch.pi * n / (N - 1)))

    @staticmethod
    def _make_hann2d(
        W: int,
        H: int,
        device: torch.device | str | None = None,
        dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        """Create a 2D Hann window."""
        device = torch.device(device) if device is not None else torch.device("cpu")
        dtype = dtype if dtype is not None else torch.float32
        hann_x = PowerSpectrumError._hann_window_1d(W, device=device, dtype=dtype)
        hann_y = PowerSpectrumError._hann_window_1d(H, device=device, dtype=dtype)
        return hann_y.view(H, 1) * hann_x.view(1, W)

    def _ensure_kgrid_and_bins(
        self,
        W: int,
        H: int,
        device: torch.device | str | None = None,
        dtype: torch.dtype | None = None,
    ) -> None:
        """Prepare frequency grid, bin edges, and pixel-to-bin-assignment."""
        if (
            (self._H, self._W) == (H, W)
            and self._k_edges is not None
            and self._bin_index is not None
        ):
            return

        device = torch.device(device) if device is not None else self.device
        dtype = dtype if dtype is not None else self.dtype

        # Frequency axis arrays
        fx = torch.fft.fftfreq(W, d=self.pixel_size, device=device, dtype=dtype)
        fy = torch.fft.fftfreq(H, d=self.pixel_size, device=device, dtype=dtype)
        if self.k_mode == "angular":
            fx = 2.0 * torch.pi * fx
            fy = 2.0 * torch.pi * fy

        # Default kmin and kmax
        kx = fx.view(1, W).expand(H, W)
        ky = fy.view(H, 1).expand(H, W)
        krad = torch.sqrt(kx * kx + ky * ky)
        kflat = krad.flatten()
        kpos = kflat[kflat > 0]
        kmin_eff = (
            self.kmin if self.kmin is not None else kpos.min().item() if kpos.numel() > 0 else 0.0
        )
        kmax_grid = float(
            torch.max(
                torch.sqrt(fx.max().abs() ** 2 + fy.max().abs() ** 2),
                torch.sqrt(fx.min().abs() ** 2 + fy.min().abs() ** 2),
            )
        )
        kmax_eff = self.kmax if self.kmax is not None else kmax_grid
        if self.log_bins:
            kmin_eff = max(kmin_eff, self.eps)

        # Bin edges
        if self.log_bins:
            edges = torch.logspace(
                torch.log10(torch.tensor(kmin_eff, device=device, dtype=dtype)),
                torch.log10(torch.tensor(kmax_eff, device=device, dtype=dtype)),
                steps=self.nbins + 1,
                base=10.0,
            )
        else:
            edges = torch.linspace(
                torch.tensor(kmin_eff, device=device, dtype=dtype),
                torch.tensor(kmax_eff, device=device, dtype=dtype),
                steps=self.nbins + 1,
            )

        # Assign each pixel to a bin: idx in [0, nbins-1], or -1 if out of range
        idx = torch.bucketize(kflat, edges) - 1
        valid = (idx >= 0) & (idx < self.nbins)
        idx = torch.where(valid, idx, torch.full_like(idx, -1))

        # Precompute counts per bin for mean calculation
        counts = torch.bincount(idx.clamp_min(0), minlength=self.nbins).to(dtype)
        counts = torch.where(counts > 0, counts, torch.ones_like(counts))

        # Save caches
        self._H, self._W = H, W
        self._k_edges = edges
        self._dk = (edges[1:] - edges[:-1]).clamp_min(self.eps)
        self._bin_index = idx
        self._counts = counts

        # Prepare Hann window if requested
        if self.window == "hann":
            self._hann2d = self._make_hann2d(H, W, device=device, dtype=dtype)
        else:
            self._hann2d = None

    def _radial_mean_power(self, power: torch.Tensor) -> torch.Tensor:
        """Compute radial mean of power given precomputed bin assignments.

        Args:
            power: power map for a single sample of shape (H, W).

        Returns:
            curve: mean power per radial bin of shape (nbins,).
        """
        assert self._bin_index is not None and self._counts is not None
        p_flat = power.flatten()
        idx = self._bin_index
        valid_mask = idx >= 0
        # Accumulate sums per bin
        sums = torch.bincount(idx[valid_mask], weights=p_flat[valid_mask], minlength=self.nbins).to(
            power.dtype
        )
        means = sums / self._counts
        return means

    @torch.no_grad()
    def update(self, data: torch.Tensor, prediction: torch.Tensor) -> None:
        """Accumulate batch power spectra errors.

        Args:
            data: Target maps of shape (B, H, W), (B, C, H, W) or (H, W).
            prediction: Predicted maps of matching shape.
        """
        targ_ = _sanitize_ndim(data)
        pred_ = _sanitize_ndim(prediction)
        if pred_.shape != targ_.shape:
            raise ValueError(f"Input shapes must match, got {targ_.shape} vs {pred_.shape}.")
        B, H, W = targ_.shape

        self._ensure_kgrid_and_bins(H, W, device=targ_.device, dtype=targ_.dtype)

        # Process each sample in the batch
        for i in range(B):
            Ip = pred_[i]
            It = targ_[i]
            if self.detrend:
                Ip = Ip - Ip.mean()
                It = It - It.mean()
            # Windowing
            if self._hann2d is not None:
                Ip = Ip * self._hann2d
                It = It * self._hann2d
            # FFT
            Fp = torch.fft.fft2(Ip, norm="ortho")
            Ft = torch.fft.fft2(It, norm="ortho")
            Pp = Fp.real.pow(2) + Fp.imag.pow(2)
            Pt = Ft.real.pow(2) + Ft.imag.pow(2)
            # Radial mean curves
            curve_p = self._radial_mean_power(Pp)
            curve_t = self._radial_mean_power(Pt)
            if self.log_power:
                curve_p = torch.log10(curve_p + self.eps)
                curve_t = torch.log10(curve_t + self.eps)
            curve_p = curve_p / curve_p.sum().clamp_min(self.eps)
            curve_t = curve_t / curve_t.sum().clamp_min(self.eps)
            per_bin_err = (curve_p - curve_t).abs()
            if self.per_bin_weighted:
                per_bin_err = per_bin_err * self._dk
            self.aggregate += per_bin_err
            self.lsq_aggregate += per_bin_err.pow(2)
            self.min_aggregate = torch.min(self.min_aggregate, per_bin_err.amin())
            self.max_aggregate = torch.max(self.max_aggregate, per_bin_err.amax())
            self.n_observations += 1

    @property
    def mean_per_bin(self) -> torch.Tensor:
        """Return the mean per-bin error curve."""
        if self.n_observations == 0:
            return self.aggregate
        return self.aggregate / float(self.n_observations)

    @property
    def var_per_bin(self) -> torch.Tensor:
        """Error variance per bin."""
        return (self.lsq_aggregate / self.n_observations) - self.mean_per_bin.pow(2)

    @property
    def std_per_bin(self) -> torch.Tensor:
        """Error variance per bin."""
        return self.var_per_bin.sqrt() / self.n_observations

    @torch.no_grad()
    def compute(self, reduction: Callable | None = None) -> torch.Tensor:
        """Return the radial profile curve error reduced to a scalar."""
        if reduction is None:
            reduction = self.reduction
        return reduction(self.mean_per_bin)



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Test configuration."""

import pytest
import matplotlib


@pytest.fixture(scope="session", autouse=True)
def show_plots(request) -> bool:
    """Check if the user wants to display plots."""
    verbose = request.config.getoption("verbose")
    return verbose >= 1



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.utils module."""

import pytest
import re
import uuid
import datetime
import numpy as np
from matplotlib.colors import Normalize, LinearSegmentedColormap
from matplotlib import pyplot as plt
import skais_mapper
from skais_mapper.utils import (
    get_run_id,
    set_run_id,
    current_time,
    compress_encode,
    extract_decode,
    alias_kw,
    primes,
    colors,
)


def test_get_set_run_id(monkeypatch):
    """Test the get_run_id and set_run_id functions."""
    # Save original
    orig_uid = getattr(skais_mapper, "RUN_UID", None)
    # Test auto-set
    set_run_id()
    uid1 = skais_mapper.RUN_UID
    assert isinstance(uid1, uuid.UUID)
    # Test set by uuid
    new_id = uuid.uuid4()
    set_run_id(new_id)
    assert skais_mapper.RUN_UID == new_id
    # Test set by string
    set_run_id(str(new_id))
    assert skais_mapper.RUN_UID == new_id
    # Test get_run_id returns correct length and hex chars
    rid = get_run_id(12)
    assert isinstance(rid, str)
    assert len(rid) == 12
    assert re.fullmatch(r"[0-9a-fA-F]+", rid)
    # Restore
    skais_mapper.RUN_UID = orig_uid


def test_current_time_variants():
    """Test the current_time function with various parameters."""
    t = current_time()
    assert isinstance(t, str)
    assert len(t) == 8
    # Test with delimiters and datetime
    t2 = current_time(date_only=True, as_str=True, no_delim=False)
    assert "-" in t2
    t3 = current_time(date_only=False, as_str=False)
    assert isinstance(t3, (datetime.date, datetime.datetime))


def test_compress_encode_and_extract_decode():
    """Test the compress_encode and extract_decode functions."""
    s = "hello world 123"
    enc = compress_encode(s)
    assert isinstance(enc, str)
    dec = extract_decode(enc)
    assert dec == s


def test_alias_kw_decorator():
    """Test the alias_kw decorator."""

    @alias_kw("foo", "bar")
    def f(foo=0):
        return foo

    assert f(foo=1) == 1
    assert f(bar=2) == 2  # Should alias bar->foo
    assert f(foo=3, bar=4) == 4  # bar should override foo


def test_legendre_basic():
    """Test the Legendre symbol calculation."""
    # (2|7) == 1, since 2 is a quadratic residue mod 7
    assert primes._legendre(2, 7) in (1, 6)  # 6 instead of -1 per doc


def test_is_sprp_true_false():
    """Test the _is_sprp function for known primes and composites."""
    # 7 is prime, 9 is not
    assert primes._is_sprp(7, 2)
    assert not primes._is_sprp(9, 2)
    assert primes._is_sprp(2503, 2)
    assert not primes._is_sprp(2509, 5)
    assert not primes._is_sprp(7801, 9)
    assert not primes._is_sprp(341, 2)
    assert primes._is_sprp(461, 2)


def test_is_lucas_prp_edge():
    """Test the _is_lucas_prp function for edge cases."""
    # For known prime, should return True
    assert primes._is_lucas_prp(7, 5)
    assert primes._is_lucas_prp(388, 2)


def test_is_prime_true_false():
    """Test the _is_sprp function for known primes and composites."""
    # 7 is prime, 9 is not
    assert primes.is_prime(7)
    assert not primes.is_prime(9)
    assert primes.is_prime(2503)
    assert not primes.is_prime(2509)
    assert not primes.is_prime(7801)
    assert primes.is_prime(1997)
    assert primes.is_prime(4211)
    assert not primes.is_prime(49729)
    assert not primes.is_prime(2147483671)
    assert primes.is_prime(2147483869)
    assert not primes.is_prime(2147483661)
    assert primes.is_prime(2147483647)


def test_next_prime():
    """Test the next_prime function."""
    assert primes.next_prime(1) == 2
    assert primes.next_prime(7) == 11
    assert primes.next_prime(40) == 41
    assert primes.next_prime(2508) == 2521
    assert primes.next_prime(2509) == 2521
    assert primes.next_prime(2511) == 2521


def test_color_variant_hex_and_shift():
    """Test the color_variant function with hex colors and shift."""
    orig = "#0080ff"
    v = colors.color_variant(orig, shift=10)
    assert v.startswith("#") and len(v) == 6
    # Should raise if not hex
    with pytest.raises(ValueError):
        colors.color_variant("0080ff")
    with pytest.raises(ValueError):
        colors.color_variant("#0080f")  # Too short


def test_color_variant_limits():
    """Test the color_variant function with limits."""
    # Lower bound
    assert colors.color_variant("#000000", shift=-10) == "#000000"
    # Upper bound
    assert colors.color_variant("#ffffff", shift=10) == "#ffffff"
    # Middle value
    assert colors.color_variant("#808080", shift=10).startswith("#")


def test_renorm_colormap_adaptor_basic():
    """Test the ReNormColormapAdaptor class."""
    cmap = plt.get_cmap('viridis')
    scmap = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize(0, 1))
    adaptor = colors.ReNormColormapAdaptor(
        scmap, lambda x: x - x.min() / (x.max() - x.min()))
    x = np.linspace(0, 1, 10)
    result = adaptor(x)
    assert isinstance(result, np.ndarray)
    # __getattr__ dispatch (grab any attribute from base)
    assert hasattr(adaptor, "N")
    # Abstract methods: just call to check present (no-op)
    adaptor._init()
    adaptor.reversed()


def test_renorm_colormap_adaptor_orig_norm():
    """Test the ReNormColormapAdaptor class."""
    cmap = plt.get_cmap('viridis')
    norm = Normalize(0, 1)
    adaptor = colors.ReNormColormapAdaptor(cmap, Normalize)
    adaptor = colors.ReNormColormapAdaptor(cmap, norm, orig_norm=Normalize(0, 0.8))
    x = np.linspace(0, 1, 10)
    result = adaptor(x)
    assert isinstance(result, np.ndarray)
    # __getattr__ dispatch (grab any attribute from base)
    assert hasattr(adaptor, "N")
    # Abstract methods: just call to check present (no-op)
    adaptor._init()
    adaptor.reversed()


def test_skaiscolors_shades_exist():
    """Test that the SkaisColors class has valid color attributes."""
    for name in ("white", "gray", "grey", "darkish"):
        val = getattr(colors.SkaisColors, name)
        assert isinstance(val, str) and val.startswith("#") and len(val) == 7


def test_skaiscolors_palette_and_random():
    """Test the SkaisColors palette and random color generation."""
    # Palette
    palette = colors.SkaisColors.palettes["hertzsprung"]
    assert isinstance(palette, list) and all(x.startswith("#") for x in palette)
    assert palette == colors.SkaisColors.hertzsprung_palette
    # Random
    cmap = colors.SkaisColors.cmap_from_color(palette[0], palette[-1])
    assert isinstance(cmap, LinearSegmentedColormap)
    cmap_native = colors.SkaisColors.cmap_from_color("cyan", "blue")
    assert isinstance(cmap_native, LinearSegmentedColormap)
    cmap_single = colors.SkaisColors.cmap_from_color("red")
    assert isinstance(cmap_single, LinearSegmentedColormap)


def test_skaiscolormaps_reverse_and_bad_values():
    """Test the reverse function and bad/over/under color settings."""
    orig = colors.SkaisColorMaps.random()
    reversed_cmap = colors.SkaisColorMaps.reverse(
        orig, set_bad="#000000", set_under="#111111", set_over="#222222"
    )
    assert isinstance(reversed_cmap, LinearSegmentedColormap)
    # Check bad/over/under colors set
    assert tuple(reversed_cmap.get_bad()) == (0.0, 0.0, 0.0, 1.0)
    assert tuple(reversed_cmap.get_over()) == (
        0.13333333333333333,
        0.13333333333333333,
        0.13333333333333333,
        1.0,
    )


def test_skaiscolormaps_reverse_and_no_bad_values():
    """Test the reverse function and bad/over/under color settings."""
    orig = colors.SkaisColorMaps.random()
    reversed_cmap = colors.SkaisColorMaps.reverse(orig)
    assert isinstance(reversed_cmap, LinearSegmentedColormap)


def test_skaiscolormaps_palette():
    """Test the SkaisColorMaps palette function."""
    interp = colors.SkaisColorMaps.palette("hertzsprung", 50)
    assert isinstance(interp, list)
    assert len(interp) == 50


def test_skaiscolormaps_as_generator():
    """Test the reverse function and bad/over/under color settings."""
    cmap = next(colors.SkaisColorMaps.gen())
    assert isinstance(cmap, LinearSegmentedColormap)


def test_register_all_runs():
    """Test the SkaisColorMaps.register_all function."""
    colors.SkaisColorMaps.register_all(verbose=True)
    cmap = plt.get_cmap("hertzsprung")
    assert isinstance(cmap, LinearSegmentedColormap)


def test_register_all_runs_non_verbose():
    """Test the SkaisColorMaps.register_all function."""
    colors.SkaisColorMaps.register_all(verbose=False)
    cmap = plt.get_cmap("hertzsprung")
    assert isinstance(cmap, LinearSegmentedColormap)


def test_plot_gradients_runs(tmp_path, monkeypatch):
    """Test the SkaisColors.plot_gradients function."""
    # Patch plt.savefig to avoid actual file creation
    import matplotlib.pyplot as plt

    called = {}

    def fake_savefig(fname):
        called["saved"] = fname

    monkeypatch.setattr(plt, "savefig", fake_savefig)
    colors.SkaisColorMaps.plot_gradients(savefig=True)
    assert "saved" in called
    called = {}
    colors.SkaisColorMaps.plot_gradients(savefig=False)
    assert "saved" not in called



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.simobjects module."""

import pytest
import numpy as np
from unittest.mock import patch, MagicMock
import astropy.units as au

import skais_mapper.simobjects as so


@pytest.fixture
def mock_tng():
    """Mock the illustris submodule."""
    with patch("skais_mapper.simobjects.tng") as tng:
        # Mock pidx/ptype conversions
        tng.util.pidx_from_ptype.side_effect = lambda x: {"gas": 0, "dm": 1, "stars": 3}.get(x, 0)
        tng.util.ptype_from_pidx.side_effect = lambda x: {0: "gas", 1: "dm", 3: "stars"}.get(
            x, "gas"
        )
        # Mock groupcat and snapshots loaders
        tng.groupcat.load_header.return_value = {
            "OmegaLambda": 0.6911,
            "Omega0": 0.3089,
            "Redshift": 0.1,
            "HubbleParam": 0.6774,
        }
        tng.snapshots.load_header.return_value = {
            "OmegaLambda": 0.6911,
            "Omega0": 0.3089,
            "Redshift": 0.1,
            "HubbleParam": 0.6774,
            "UnitLength_in_cm": 3.0e21,
            "UnitMass_in_g": 1.0e43,
            "UnitVelocity_in_cm_per_s": 1e5,
            "MassTable": [0, 3.07368e-05, 0, 5.73879e-06, 0, 0],
        }
        tng.groupcat.load_halos.return_value = np.array([0, 1, 2])
        tng.groupcat.load_single.return_value = {
            "SubhaloLen": 10,
            "SubhaloLenType": [1, 2, 3],
            "SubhaloPos": np.array([1, 1, 1]),
            "SubhaloHalfmassRad": np.ones(1),
            "SubhaloHalfmassRadType": np.ones(1),
        }
        tng.snapshots.snapshot_offsets.return_value = {
            "offsetType": np.zeros((6,)),
            "snapOffsets": np.zeros((6, 2)),
            "lenType": np.zeros((6,)),
        }
        tng.snapshots.load_snapshot.return_value = {
            "Coordinates": np.zeros((3, 3)),
            "Redshift": np.zeros(3),
            "Masses": np.ones(3),
            "Velocities": np.zeros((3, 3)),
            "MagneticField": np.ones((3, 3)),
            "Density": np.ones(3),
            "SubfindDensity": np.ones(3),
            "SubfindDMDensity": np.ones(3),
            "InternalEnergy": np.ones(3),
            "ElectronAbundance": np.ones(3),
            "NeutralHydrogenAbundance": np.ones(3),
            "GFM_Metals": np.ones((3, 10)),
            "CenterOfMass": np.zeros((3, 3)),
        }
        yield tng


@pytest.fixture
def galaxy(mock_tng):
    """Mock the `TNGGalaxy` class, including a `CosmoModel`."""
    with patch("skais_mapper.simobjects.CosmoModel") as Cosmo:
        cosmo = MagicMock()
        cosmo.h = 0.7
        cosmo.a = 1.0
        cosmo.z = 0.5
        cosmo.d_z.return_value = 1
        cosmo.d_z2kpc.return_value = au.kpc
        cosmo.arcsec2kpc.return_value = au.kpc
        Cosmo.return_value = cosmo
        g = so.TNGGalaxy("/tmp", 0, 0, as_float32=False)
        g.data.update({"GFM_Metals": np.array([[0.76]]), "NeutralHydrogenAbundance": np.ones(3)})
        return g


def test_sphgalaxy_init_sets_attributes(galaxy):
    """Test `SPHGalaxy` initialization sets attributes correctly."""
    assert galaxy.ra.unit == au.deg
    assert galaxy.dec.unit == au.deg
    assert galaxy.distance.unit == au.Mpc
    assert galaxy.peculiar_v.unit == au.km / au.s
    assert hasattr(galaxy, "header")
    assert hasattr(galaxy, "cosmology")
    assert hasattr(galaxy, "data")


def test_sphgalaxy_basic_paths(mock_tng):
    """Test `SPHGalaxy` unit and property paths not hit by `TNGGalaxy`."""
    g = so.SPHGalaxy()
    # Test p_idx setter for int and str
    g.p_idx = 1
    assert g._p_idx == 1
    g.p_idx = "gas"
    assert isinstance(g._p_idx, int)
    # Test particle_type setter
    g.particle_type = "dm"
    assert isinstance(g.p_idx, int)
    # Test set_units and odd unit strings
    g.set_units(length=1e21, velocity=1e5, mass=1e40)
    assert isinstance(g.units("t"), au.Quantity)
    assert isinstance(g.units("sqrtP"), au.Quantity)
    assert g.units("unknown") == 1
    # Test float32 paths
    g.as_float32 = True
    assert g.UnitLength.dtype == np.float32 or hasattr(g.UnitLength, "dtype")
    assert g.UnitMass.dtype == np.float32 or hasattr(g.UnitMass, "dtype")
    assert g.UnitVelocity.dtype == np.float32 or hasattr(g.UnitVelocity, "dtype")


def test_sphgalaxy_particle_type_none_sets_default(mock_tng):
    """Test `SPHGalaxy` with no particle type set, should default to 'gas'."""
    g = so.SPHGalaxy(particle_type=None)
    assert g._p_idx == 0  # default 'gas'


def test_galaxy_explicit_type_and_unit(mock_tng):
    """Test `TNGGalaxy` with non-fallback particle type and units."""
    g = so.TNGGalaxy(
        "/tmp", 0, 0, particle_type="dm", units={"length": 1, "velocity": 1, "mass": 1}
    )
    assert g.particle_type == "dm"
    assert g.p_idx == 1


def test_sphgalaxy_units_setter(monkeypatch, mock_tng):
    """Test `SPHGalaxy` units setter."""
    g = so.SPHGalaxy(units={"length": "kpc", "mass": "Msun", "velocity": "km/s"})
    # If set_units is called, it should set the attributes
    g.set_units = MagicMock()
    g.__init__(units={"length": "kpc", "mass": "Msun", "velocity": "km/s"})
    g.set_units.assert_called_with(length="kpc", mass="Msun", velocity="km/s")


def test_galaxy_p_idx_setter(mock_tng):
    """Test `SPHGalaxy` particle index getting/setting."""
    g = so.SPHGalaxy()
    g.p_idx = "gas"
    assert g.p_idx == 0
    g.p_idx = 1
    assert g.particle_type == "dm"
    g.p_idx = None
    assert g.particle_type == "gas"


def test_sphgalaxy_pidx_property_and_setter(galaxy):
    """Test `SPHGalaxy.p_idx` property and setter."""
    # Test default
    assert isinstance(galaxy.p_idx, int)
    # Test setter with int
    galaxy._p_idx = 0
    galaxy.p_idx = "dm"
    assert galaxy._p_idx == 1
    # Test setter with string
    with patch("skais_mapper.simobjects.tng.util.pidx_from_ptype", return_value=4) as mock_pidx:
        galaxy.p_idx = "star"
        mock_pidx.assert_called_with("star")
        assert galaxy._p_idx == 4


def test_particle_mass_and_other_data_fields(mock_tng):
    """Test `SPHGalaxy.particle_mass` and related data-dependent properties."""
    g = so.SPHGalaxy()
    # Add MassTable for constant particle mass
    g.header["snapshot"]["MassTable"] = [0, 1, 2, 3, 4, 5]
    g._p_idx = 1
    pm = g.particle_mass
    assert pm is None or isinstance(pm, au.Quantity)
    # Populate data fields for other accessors
    g.data = {
        "CenterOfMass": np.zeros((3, 3)),
        "Velocities": np.ones((3, 3)),
        "InternalEnergy": np.ones(3),
        "ElectronAbundance": np.ones(3),
        "GFM_Metals": np.ones((3, 1)),
        "NeutralHydrogenAbundance": np.ones(3),
    }
    assert g.cell_positions.shape[0] == 3
    assert g.velocities.shape[0] == 3
    assert g.internal_energy.shape[0] == 3
    assert g.x_e.shape[0] == 3
    assert g.x_H.shape[0] == 3
    assert g.x_HI.shape[0] == 3


def test_sphgalaxy_rotation_and_verbose_flags(galaxy):
    """Test `SPHGalaxy` rotation and verbose flags."""
    assert hasattr(galaxy, "rotation")
    assert hasattr(galaxy, "verbose")
    # Check that verbose flag is set correctly
    g = galaxy
    g.verbose = True
    assert g.verbose is True


def test_load_cosmology(mock_tng):
    """Test `SPHGalaxy` particle index getting/setting."""
    g = so.SPHGalaxy()
    g.load_cosmology(
        {
            "omega_l": 0.7,
            "omega_k": 0.3,
            "z": 0,
            "h": 0.7,
        }
    )
    assert g.cosmology.z == 0
    assert g.cosmology.h == 0.7
    g.load_cosmology(None, in_place=False)
    assert g.cosmology.z == 0
    assert g.cosmology.h == 0.7


def test_sphgalaxy_init_with_all_args(mock_tng):
    """Test `SPHGalaxy` initialization with all arguments."""
    rotation = lambda x: x
    units = {"length": "kpc", "mass": "Msun", "velocity": "km/s"}
    g = so.SPHGalaxy(
        ra=10,
        dec=20,
        distance=5,
        peculiar_v=100,
        rotation=rotation,
        cosmo_pars={"omega_k": 0.3},
        units=units,
        particle_type="dm",
        as_float32=True,
        verbose=True,
    )
    assert g.ra.value == 10
    assert g.dec.value == 20
    assert g.distance.value == 5
    assert g.peculiar_v.value == 100
    assert g.rotation is rotation
    assert g.as_float32 is True
    assert g.verbose is True


def test_load_tng_cosmology(mock_tng):
    """Test `TNGGalaxy` particle index getting/setting."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.load_cosmology(
        {
            "omega_l": 0.7,
            "omega_k": 0.3,
            "z": 0,
            "h": 0.721,
        },
        in_place=True,
    )

    assert g.cosmology.omega_l == 0.7
    assert g.cosmology.omega_k == 0.3
    assert g.cosmology.z == 0
    assert g.cosmology.h == 0.721
    g.load_cosmology(None, in_place=True)
    assert g.cosmology.z == 0.1
    assert g.cosmology.h == 0.6774
    g.load_cosmology(
        {
            "omega_l": 0.7,
            "omega_k": 0.3,
            "z": 0,
            "h": 0.721,
        },
        in_place=False,
    )
    assert g.cosmology.z == 0.1
    assert g.cosmology.h == 0.6774


def test_cosmo_angular_dist_and_res(mock_tng):
    """Test `SPHGalaxy` particle index getting/setting."""
    g = so.SPHGalaxy()
    g.load_cosmology({"z": 0})
    dist_ang = g.angular_distance(eps=1e-3)
    assert dist_ang >= 0
    res_ang = g.angular_resolution(eps=1e-3)
    assert res_ang[0] >= 0


def test_load_data_and_header(mock_tng):
    """Force load_data and other methods to trigger fallback loading header."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.load_data(
        particle_type="gas",
        primary_fields=["Coordinates", "Masses"],
        optional_fields=["Density"],
        verbose=True,
    )
    assert isinstance(g.data, dict)


def test_load_data_gfm_metals_and_com():
    """Force load_data and other methods to trigger fallback loading header."""
    with patch("skais_mapper.simobjects.tng") as tng:
        tng.util.pidx_from_ptype.side_effect = lambda x: {"gas": 0, "dm": 1, "stars": 3}.get(x, 0)
        tng.util.ptype_from_pidx.side_effect = lambda x: {0: "gas", 1: "dm", 3: "stars"}.get(
            x, "gas"
        )
        # Mock groupcat and snapshots loaders
        tng.groupcat.load_header.return_value = {
            "OmegaLambda": 0.6911,
            "Omega0": 0.3089,
            "Redshift": 0.1,
            "HubbleParam": 0.6774,
        }
        tng.snapshots.load_header.return_value = {
            "OmegaLambda": 0.6911,
            "Omega0": 0.3089,
            "Redshift": 0.1,
            "HubbleParam": 0.6774,
            "UnitLength_in_cm": 3.0e21,
            "UnitMass_in_g": 1.0e43,
            "UnitVelocity_in_cm_per_s": 1e5,
            "MassTable": [0, 3.07368e-05, 0, 5.73879e-06, 0, 0],
        }
        tng.groupcat.load_halos.return_value = np.array([0, 1, 2])
        tng.groupcat.load_single.return_value = {
            "SubhaloLen": 10,
            "SubhaloLenType": [1, 2, 3],
            "SubhaloPos": np.array([1, 1, 1]),
            "SubhaloHalfmassRad": np.ones(1),
            "SubhaloHalfmassRadType": np.ones(1),
        }
        tng.snapshots.snapshot_offsets.return_value = {
            "offsetType": np.zeros((6,)),
            "snapOffsets": np.zeros((6, 2)),
            "lenType": np.zeros((6,)),
        }
        tng.snapshots.load_snapshot.return_value = {
            "Coordinates": np.zeros((3, 3)),
            "Redshift": np.zeros(3),
            "Masses": np.ones(3),
            "Velocities": np.zeros((3, 3)),
            "Density": np.ones(3),
            "SubfindDensity": np.ones(3),
            "SubfindDMDensity": np.ones(3),
            "InternalEnergy": np.ones(3),
            "ElectronAbundance": np.ones(3),
            "NeutralHydrogenAbundance": np.ones(3),
            "GFM_Metals": None,
            "CenterOfMass": None,
        }

        g = so.TNGGalaxy("/tmp", 0, 0)
        g.load_data(
            particle_type="gas",
            primary_fields=["Coordinates", "Masses"],
            verbose=True,
        )
    assert isinstance(g.data, dict)


def test_load_data_density_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Density
    del g.data
    assert len(g.density) == 3
    del g.data["Density"]
    assert len(g.density) == 3
    del g.data["SubfindDMDensity"]
    assert len(g.density) == 3
    del g.data["SubfindDensity"]
    assert g.density is None


def test_load_data_masses_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Masses
    del g.data
    assert len(g.masses) == 3
    del g.data["Masses"]
    assert len(g.masses) == 3
    del g.header["snapshot"]["MassTable"]
    assert g.masses is None


def test_load_data_internal_energy_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Internal Energy
    del g.data
    assert len(g.internal_energy) == 3
    del g.data["InternalEnergy"]
    assert g.internal_energy is None


def test_load_data_electron_abundance_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Electron abundance
    del g.data
    assert len(g.x_e) == 3
    del g.data["ElectronAbundance"]
    assert g.x_e is None


def test_load_data_x_h_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Hydrogen
    del g.data
    assert len(g.x_H) == 3
    del g.data["GFM_Metals"]
    assert g.x_H is None


def test_load_data_n_h_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Hydrogen number density
    del g.data
    assert len(g.n_H) == 3
    del g.data["Density"]
    del g.data["SubfindDMDensity"]
    del g.data["SubfindDensity"]
    assert g.n_H is None


def test_load_data_m_h_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Hydrogen density
    del g.data
    assert len(g.m_H) == 3
    del g.data["Masses"]
    del g.header["snapshot"]["MassTable"]
    assert g.m_H is None


def test_load_data_x_hi_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Neutral Hydrogen
    del g.data
    assert len(g.x_HI) == 3
    del g.data["NeutralHydrogenAbundance"]
    assert g.x_HI is None


def test_load_data_n_hi_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Hydrogen number density
    del g.data
    assert len(g.n_HI) == 3
    del g.data["Density"]
    del g.data["SubfindDMDensity"]
    del g.data["SubfindDensity"]
    del g.data["NeutralHydrogenAbundance"]
    assert g.n_HI is None


def test_load_data_m_hi_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Hydrogen density
    del g.data
    assert len(g.m_HI) == 3
    del g.data["Masses"]
    del g.header["snapshot"]["MassTable"]
    assert g.m_HI is None


def test_load_data_magnetic_field_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Magnetic Field
    del g.data
    assert len(g.magnetic_field) == 3
    del g.data["MagneticField"]
    assert g.magnetic_field is None


def test_load_data_velocities_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Velocities
    del g.data
    assert len(g.velocities) == 3
    del g.data["Velocities"]
    assert g.velocities is None


def test_load_data_positions_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    g.p_idx = 1
    # Positions
    del g.data
    assert len(g.particle_positions) == 3
    del g.data["Coordinates"]
    assert g.particle_positions is None
    del g.data
    assert len(g.cell_positions) == 3
    del g.data["CenterOfMass"]
    assert g.cell_positions is None


def test_load_header_property_fallback(mock_tng):
    """Force load_header as property access fallback."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    del g.header
    bs = g.boxsize
    L = g.units("l/h")
    assert g.units("l") / g.cosmology.h * g.cosmology.a == L
    assert (bs / L).decompose() == 1
    # M
    M = g.units("m")
    assert M.value == 1e43
    del g.header
    g.set_units(1, 1, 1)
    del g.header
    M = g.UnitMass
    assert M.value == 1
    del g.header["snapshot"]["UnitMass_in_g"]
    M = g.UnitMass
    assert M.value == 1.989e43
    # L
    del g.header
    L = g.UnitLength
    assert L.value == 1
    del g.header["snapshot"]["UnitLength_in_cm"]
    L = g.UnitLength
    assert L.value == 3.085678e21
    # V
    del g.header
    V = g.UnitVelocity
    assert V.value == 1
    del g.header["snapshot"]["UnitVelocity_in_cm_per_s"]
    L = g.UnitVelocity
    assert L.value == 1e5


def test_subhalo_list_and_filtering(mock_tng):
    """Exercise subhalo_list with filtering by mass and radius."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    # Default (no filter) returns array
    all_ids = g.subhalo_list("/tmp", 0, verbose=True)
    assert isinstance(all_ids, np.ndarray)
    # Filter by mass and radius, should still return array (mock data has small values)
    filtered_ids = g.subhalo_list("/tmp", 0, filtered=True)
    assert isinstance(filtered_ids, np.ndarray)
    del g._halo_index
    assert g.halo_index == 0


def test_subhalo_properties(mock_tng):
    """Exercise `TNGGalaxy` properties depending on subhalo."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    print(g.subhalo)


def test_halo_indices(mock_tng):
    """Exercise `TNGGalaxy.halo_index` manipulations."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    del g._halo_index
    assert g.halo_index == 0
    g._halo_index = 0
    g.halo_index = 0
    g.subhalo = g.load_subhalo(0)
    g.data = g.load_data(0)
    g.halo_index = 1
    assert g.subhalo is not None
    assert g.data is not None
    g.subhalo = g.load_subhalo(-1, verbose=True)


def test_units_and_properties(galaxy):
    """Test `TNGGalaxy.Unit*` properties."""
    assert galaxy.UnitLength.unit == au.cm
    assert galaxy.UnitMass.unit == au.g
    assert galaxy.UnitVelocity.unit == (au.cm / au.s)
    assert isinstance(galaxy.units("vp"), au.Quantity)

    # density and masses use mocked data
    d = galaxy.density
    m = galaxy.masses
    assert d.shape[0] == 3 and m.shape[0] == 3
    assert galaxy.n_H is not None
    assert galaxy.m_H is not None
    assert galaxy.n_HI is not None
    assert galaxy.m_HI is not None


def test_no_subhalo_defaults(mock_tng):
    """Test `TNGGalaxy.N_particles_type` without subhalo."""
    g = so.TNGGalaxy("/tmp", 0, 0)
    assert g.N_particles == 10
    assert len(g.N_particles_type) == 3
    del g.subhalo
    assert len(g.N_particles_type) == 0
    assert g.N_particles == -1
    # center loads subhalo anew
    assert len(g.center) == 3
    del g.subhalo["SubhaloPos"]
    assert g.center is None


def test_center_radii_hsml_properties(galaxy):
    """Ensure center, radii, and hsml properties resolve from mocked subhalo fields."""
    center = galaxy.center
    radii = galaxy.radii
    hsml = galaxy.hsml
    assert center.shape[0] == 3
    assert radii.shape[0] == 1 or radii.shape[0] == 3
    assert hsml.shape[0] >= 1


def test_radii_temperature_magnetic(galaxy):
    """Test `TNGGalaxy.{radii, hsml, temperature, magnetic_field_strength}` properties."""
    galaxy.data["MagneticField"] = np.ones((3, 3))
    T = galaxy.temperature
    assert T.unit == au.K
    b = galaxy.magnetic_field_strength
    assert b.unit.is_equivalent(au.G)
    assert galaxy.radii.shape[0] == 3
    assert galaxy.hsml.shape[0] == 3


def test_kd_tree_returns_array(galaxy):
    """Test `TNGGalaxy.kd_tree` case: returns array."""
    galaxy.data["Coordinates"] = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])
    result = galaxy.kd_tree(k=2, threads=1, verbose=True)
    assert result.shape[0] == 4


def test_kd_tree_returns_None(galaxy):
    """Test `TNGGalaxy.kd_tree` case: returns None."""
    del galaxy.data["Coordinates"]
    result = galaxy.kd_tree(k=2, threads=1, verbose=False)
    assert result is None
    galaxy.data["Coordinates"] = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1]])


def test_kdtree_and_strip_units_variants(galaxy):
    """Cover kd_tree float32 and strip_ap_units dtype case."""
    galaxy.data["Coordinates"] = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0]], dtype=float)
    # as_float32 path
    dists = galaxy.kd_tree(k=2, threads=1, verbose=False, as_float32=True)
    assert dists.dtype == np.float32
    # strip_ap_units forcing float32 dtype
    q = np.array([1, 2, 3]) * au.kpc
    vals = so.strip_ap_units(q, mask=np.array([0, 2]), units=[au.kpc], dtype=np.float32)
    assert vals[0].dtype == np.float32


def test_get_mapping_arrays(galaxy):
    """Test `TNGGalaxy.get_mapping_arrays` method."""
    galaxy.data["Coordinates"] = np.zeros((3, 3))
    galaxy.data["Masses"] = np.ones(3)
    galaxy.data["Density"] = np.ones(3)
    galaxy.data["InternalEnergy"] = np.ones(3)
    galaxy.data["ElectronAbundance"] = np.ones(3)
    galaxy.subhalo["magnetic_field_strength"] = np.ones((3, 3))
    del galaxy.data["MagneticField"]
    arrs = galaxy.get_mapping_arrays(
        ("particle_positions", ("temperature", "masses")), verbose=True
    )
    arrs = galaxy.get_mapping_arrays(
        ("particle_positions", ("magnetic_field_strength", "magnetic_field_strength"))
    )
    print(arrs)


def test_mapping_and_map_generation(galaxy):
    """Test `TNGGalaxy.get_mapping_arrays` & `TNGGalaxy.generate_map` methods."""
    galaxy.data["Coordinates"] = np.zeros((3, 3))
    galaxy.data["Masses"] = np.ones(3)
    galaxy.data["Density"] = np.ones(3)
    galaxy.data["InternalEnergy"] = np.ones(3)
    galaxy.data["ElectronAbundance"] = np.ones(3)
    arrs = galaxy.get_mapping_arrays()
    assert len(arrs) == 3
    # test invalid assignment function triggers ValueError
    with pytest.raises(ValueError):
        galaxy.generate_map(assignment_func=lambda *a, **k: None)
    # patch voronoi_RT_2D to no-op
    with patch("skais_mapper.simobjects.voronoi_RT_2D") as mock_fn:
        proj, extent, count = galaxy.generate_map(assignment_func=mock_fn, verbose=False)
        mock_fn.assert_called_once()
        assert proj.shape[0] == proj.shape[1]


def test_generate_map_with_halfmassrad(galaxy):
    """Trigger the rotation branch in generate_map."""
    galaxy.data["Coordinates"] = np.zeros((3, 3))
    galaxy.data["Masses"] = np.ones(3)
    # Patch voronoi_RT_2D to bypass actual computation
    with patch("skais_mapper.simobjects.voronoi_RT_2D") as mock_fn:
        mock_fn.__name__ = "voronoi_RT_2D_mock"
        proj, extent, count = galaxy.generate_map(
            keys=["particle_positions", "masses", "radii", "center", "SubhaloHalfmassRadType"],
            factors=[1, 1, 1, 1, 1],
            use_half_mass_rad=True,
            grid_size=8,
            assignment_func=mock_fn,
            tracers=np.ones(3),
            divisions=np.ones(3),
            verbose=True,
        )
        mock_fn.assert_called_once()
        assert proj.shape == (8, 8)
        assert len(extent) == 4
        assert isinstance(count, int)


def test_generate_map_with_rotation(galaxy):
    """Trigger the rotation branch in generate_map."""
    galaxy.data["Coordinates"] = np.zeros((3, 3))
    galaxy.data["Masses"] = np.ones(3)
    # Patch voronoi_RT_2D to bypass actual computation
    with patch("skais_mapper.simobjects.voronoi_RT_2D") as mock_fn:
        proj, extent, count = galaxy.generate_map(
            keys=["particle_positions", "masses", "radii", "center"],
            factors=[1, 1, 1, 1],
            use_half_mass_rad=False,
            grid_size=8,
            assignment_func=mock_fn,
            rot=(1, 0, 0),
            verbose=False,
        )
        mock_fn.assert_called_once()
        assert proj.shape == (8, 8)
        assert len(extent) == 4
        assert isinstance(count, int)


def test_generate_map_with_exception(galaxy):
    """Trigger the rotation branch in generate_map."""
    galaxy.data["Coordinates"] = np.zeros((3, 3))
    galaxy.data["Masses"] = np.ones(3)
    # Patch voronoi_RT_2D to bypass actual computation
    with patch("skais_mapper.simobjects.voronoi_RT_2D") as mock_fn:
        mock_fn.side_effect = Exception("Mock exception")
        with pytest.raises(Exception):
            proj, extent, count = galaxy.generate_map(
                keys=["particle_positions", "masses", "radii", "center"],
                factors=[1, 1, 1, 1],
                use_half_mass_rad=False,
                grid_size=8,
                assignment_func=mock_fn,
                rot=(1, 0, 0),
                verbose=False,
            )


def test_indices_within_box_without_radius(capfd):
    """Test indices_within_box branch where radius is computed from extent."""
    pts = np.array([[0, 0, 0], [2, 0, 0], [0, 2, 0]]) * au.kpc
    center = np.zeros(3) * au.kpc
    idcs, limits = so.indices_within_box(pts, center, radius=None, verbose=True)
    assert isinstance(idcs, np.ndarray)
    assert all(isinstance(x, au.Quantity) for x in limits)
    out, err = capfd.readouterr()
    assert "Box extent" in out
    assert "Radius" in out
    assert "Extent" in out
    assert "Selected particles" in out


def test_indices_within_box_and_strip_units():
    """Test `indices_within_box` method."""
    pts = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]]) * au.kpc
    center = np.array([0, 0, 0]) * au.kpc
    idcs, limits = so.indices_within_box(pts, center, radius=1 * au.kpc)
    assert isinstance(idcs, np.ndarray) and len(limits) == 6
    vals = so.strip_ap_units(pts, mask=np.array([0, 1]), units=[au.kpc])
    assert isinstance(vals[0], np.ndarray)


def test_strip_units():
    """Test `strip_ap_units` function."""
    # Test with a single array
    pts = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]]) * au.kpc
    stripped = so.strip_ap_units(pts, mask=np.array([0]))
    assert isinstance(stripped[0], np.ndarray)
    # assert stripped[0].dtype == np.float64

    stripped = so.strip_ap_units(pts)
    assert isinstance(stripped[0], np.ndarray)



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.configure module."""

import pytest
import logging
from unittest.mock import MagicMock
import skais_mapper.configure as configure


@pytest.fixture
def mock_cfg():
    """Mock configuration object for testing."""
    class DummyCfg(dict):
        exclude_git_state = False
        include_git_diff = False
        verbose = False
        save_configs = False
        output = "output.yaml"
    return DummyCfg()


@pytest.fixture
def mock_opt():
    """Mock optimization object for testing."""
    class DummyOpt:
        run_id = "test_run"
        verbose = False
        save_configs = False
        output = "output.yaml"
    return DummyOpt()


@pytest.fixture
def patch_dependencies(monkeypatch, mock_opt):
    """Patch dependencies for the configure module."""
    # Patch HydraConfig.get()
    hydra_config = MagicMock()
    hydra_config.runtime.output_dir = "mock_output_dir"
    hydra_config.output_subdir = "mock_subdir"
    monkeypatch.setattr(configure.HydraConfig, "get", lambda: hydra_config)

    # Patch instantiate
    monkeypatch.setattr(configure, "instantiate", lambda cfg: mock_opt)
    # Patch compress_encode
    monkeypatch.setattr(configure, "compress_encode", lambda x: f"compressed({x})")
    # Patch skais_mapper.GIT_STATE and GIT_DIFF
    monkeypatch.setattr(configure.skais_mapper, "GIT_STATE", "FAKE_GIT_STATE")
    monkeypatch.setattr(configure.skais_mapper, "GIT_DIFF", ["diff1", "diff2"])
    # Patch OmegaConf.to_yaml
    monkeypatch.setattr(configure.OmegaConf, "to_yaml", lambda cfg: "yaml_output")
    # Patch Path.read_bytes and write_bytes
    mock_path = MagicMock()
    mock_path.read_bytes.return_value = b"cfg"
    monkeypatch.setattr(configure, "Path", lambda x=None: mock_path)


@pytest.mark.usefixtures("patch_dependencies")
def test_configure_logs_git_state_and_diff(monkeypatch, caplog, mock_cfg, mock_opt):
    """Test that create function logs git state and diff correctly."""
    mock_cfg.exclude_git_state = False
    mock_cfg.include_git_diff = True
    mock_cfg.verbose = False
    mock_cfg.save_configs = False

    with caplog.at_level(logging.INFO):
        configure.create(mock_cfg)
        assert "Git state: FAKE_GIT_STATE" in caplog.text
        assert "Git diff: compressed(diff1)" in caplog.text
        assert "Git diff: compressed(diff2)" in caplog.text
        assert "Job id: test_run" in caplog.text
        assert "Output directory: mock_output_dir" in caplog.text


@pytest.mark.usefixtures("patch_dependencies")
def test_configure_exclude_git_state(monkeypatch, caplog, mock_cfg, mock_opt):
    """Test that create function does not log git state when excluded."""
    mock_cfg.exclude_git_state = True
    mock_cfg.include_git_diff = False
    mock_cfg.verbose = False
    mock_cfg.save_configs = False

    with caplog.at_level(logging.INFO):
        configure.create(mock_cfg)
        assert "Git state:" not in caplog.text
        assert "Git diff:" not in caplog.text


@pytest.mark.usefixtures("patch_dependencies")
def test_create_prints_config(monkeypatch, mock_cfg, mock_opt, capsys):
    """Test that create function prints configuration when verbose."""
    mock_cfg.verbose = True
    mock_opt.verbose = True
    mock_cfg.save_configs = False

    configure.create(mock_cfg)
    captured = capsys.readouterr()
    assert "Configuration:" in captured.out
    assert "yaml_output" in captured.out


@pytest.mark.usefixtures("patch_dependencies")
def test_create_saves_config(monkeypatch, mock_cfg, mock_opt):
    """Test that create function saves configuration to file."""
    mock_cfg.save_configs = True
    mock_opt.save_configs = True
    mock_opt.output = "output.yaml"

    mock_path = MagicMock()
    mock_path.suffix = ".yaml"
    mock_path.read_bytes.return_value = b"cfg"
    monkeypatch.setattr(configure, "Path", lambda x=None: mock_path)

    configure.create(mock_cfg)
    # Should attempt to write config bytes
    assert mock_path.write_bytes.called


@pytest.mark.usefixtures("patch_dependencies")
def test_create_saves_config_non_yaml(monkeypatch, mock_cfg, mock_opt):
    """Test that create function saves configuration to file."""
    mock_cfg.save_configs = True
    mock_opt.save_configs = True
    mock_opt.output = "output.txt"

    mock_path = MagicMock()
    mock_path.suffix = ".txt"
    mock_path.read_bytes.return_value = b"cfg"
    monkeypatch.setattr(configure, "Path", lambda x=None: mock_path)

    configure.create(mock_cfg)
    # Should attempt to write config bytes
    assert mock_path.write_bytes.called



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.illustris.util module."""

import pytest
import types
import skais_mapper.illustris.util as util


# @pytest.fixture
# def dummy_path_func():
#     """A dummy path function that returns a fixed path for testing."""
#     return lambda base, snap, part: f"/fake/path/snap_{snap}_part_{part}.hdf5"


# @pytest.fixture
# def mock_nbytes(monkeypatch):
#     """Mock the nbytes function to return a fixed size for testing."""
#     monkeypatch.setattr(
#         util, "nbytes", lambda x: 16 if x == "16" else int(str(x).replace("G", "1000000000"))
#     )


# @pytest.fixture
# def fake_file(monkeypatch):
#     """A fake h5py File class to simulate the h5py.File context manager."""

#     class DummyH5File:
#         def __init__(self, *args, **kwargs):
#             self.args = args
#             self.kwargs = kwargs

#         def __enter__(self):
#             return self

#         def __exit__(self, *a, **k):
#             return None

#         def __str__(self):
#             return "<FakeH5File>"

#     monkeypatch.setattr(util, "h5py", types.SimpleNamespace(File=DummyH5File))
#     return DummyH5File


@pytest.mark.parametrize(
    "ptype,expected",
    [
        (0, 0),
        (1, 1),
        ("dm", 1),
        ("dmlowres", 2),
        ("gas", 0),
        ("star", 4),
        ("bh", 5),
        ("tracer", 3),
        ("wind", 4),
        ("star", 4),
    ],
)
def test_pidx_from_ptype(ptype, expected):
    """Test the pidx_from_ptype function with various particle types."""
    assert util.pidx_from_ptype(ptype) == expected


def test_pidx_from_ptype_invalid():
    """Test the pidx_from_ptype function with an invalid particle type."""
    with pytest.raises(ValueError):
        util.pidx_from_ptype("invalid")


@pytest.mark.parametrize(
    "idx,expected",
    [(0, "gas"), (1, "dm"), (2, "dmlowres"), (3, "tracer"), (4, "star"), (5, "bh"), ("dm", "dm")],
)
def test_ptype_from_pidx(idx, expected):
    """Test the ptype_from_pidx function with various particle indices."""
    assert util.ptype_from_pidx(idx) == expected


def test_ptype_from_pidx_invalid():
    """Test the ptype_from_pidx function with an invalid particle index."""
    with pytest.raises(ValueError):
        util.ptype_from_pidx(99)



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.profile module."""

from __future__ import annotations
from typing import TYPE_CHECKING
from skais_mapper._compat import TORCH_AVAILABLE
import pytest
import matplotlib.pyplot as plt
if TORCH_AVAILABLE or TYPE_CHECKING:
    import torch
    import skais_mapper.profile as profile
else:
    from skais_mapper import _torch_stub as _stub  # noqa
    from skais_mapper._torch_stub import *  # noqa: F401,F403


def generate_gaussian2d_batch(
    means: torch.Tensor,
    sigmas: torch.Tensor,
    rho: torch.Tensor | None,
    size: tuple[int, int],
    normalized: bool = True,
    device: torch.device | str | None = None,
    dtype: torch.dtype | None = None,
) -> torch.Tensor:
    """Generate a batch of 2D Gaussian distributions.

    Args:
        means: Mean values for the Gaussians of shape (B, 2).
        sigmas: Standard deviations for the Gaussians of shape (B, 2).
        rho: Correlation coefficients for the Gaussians of shape (B,).
        size: Size of the output grid (W, H).
        normalized: Whether to normalize the Gaussians.
        device: Device to place the output tensor on.
        dtype: Data type of the output tensor.
    """
    B = means.shape[0]
    W, H = size
    device = torch.device(device) if device is not None else means.device
    dtype = dtype if dtype is not None else means.dtype
    means = means.to(device=device, dtype=dtype)
    sigmas = sigmas.to(device=device, dtype=dtype)
    rho = (
        rho.to(device=device, dtype=dtype)
        if rho is not None
        else torch.zeros(B, device=device, dtype=dtype)
    )
    # Grid
    xx, yy = profile._make_grid(W, H, device=device, dtype=dtype)
    mx, my = means[:, 0][:, None, None], means[:, 1][:, None, None]
    sx, sy = sigmas[:, 0][:, None, None], sigmas[:, 1][:, None, None]
    r = rho[:, None, None].clamp(-0.999, 0.999)
    dx = xx[None, ...] - mx
    dy = yy[None, ...] - my
    xs = dx / sx
    ys = dy / sy
    denom = 2.0 * (1.0 - r**2).clamp_min(1e-12)
    exponent = -(xs**2 + ys**2 - 2.0 * r * xs * ys) / denom
    g = torch.exp(exponent)
    if normalized:
        norm = 1.0 / (2.0 * torch.pi * sx * sy * torch.sqrt((1.0 - r**2).clamp_min(1e-12)))
        g = g * norm
    return g.unsqueeze(1)


@pytest.fixture
def gaussians() -> torch.Tensor:
    """Fixture to generate a default batch of 2D Gaussians."""
    W, H = 512, 512
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    dtype = torch.float32
    means = torch.tensor([
        [256.0, 256.0],
        [256.0, 256.0],
        [256.0, 256.0],
        [256.0, 256.0],
        [256.0, 256.0],
        [128.0, 128.0],
    ], device=device, dtype=dtype)

    sigmas = torch.tensor([
        [40.0, 60.0],
        [32.0, 32.0],
        [20.0, 10.0],
        [20.0, 24.0],
        [30.0, 8.0],
        [24.0, 20.0],
    ], device=device, dtype=dtype)
    rho = torch.tensor([0.0, 0.5, -0.3, 0.0, -0.8, 0.1], device=device, dtype=dtype)
    return generate_gaussian2d_batch(
        means, sigmas, rho, size=(W, H), device=device, dtype=dtype
    )

@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
@pytest.mark.parametrize(
    "x",
    [
        torch.rand((10, 10)),
        torch.rand((1, 10, 10)),
        torch.rand((1, 1, 10, 10)),
    ],
)
def test_sanitize_ndim(x):
    """Test `sanitize_ndim` function."""
    # Test 2/3/4D input tensor
    result = profile._sanitize_ndim(x)
    assert result.shape == (1, 10, 10)


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_sanitize_ndim_invalid():
    """Test `sanitize_ndim` function: case ValueError."""
    # Test invalid dimensions
    x = torch.rand((1, 1, 10, 10, 10))
    with pytest.raises(ValueError):
        profile._sanitize_ndim(x)


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test__make_grid():
    """Test the _make_grid function."""
    X, Y = profile._make_grid(10, 10)
    assert X.shape == (10, 10)
    assert Y.shape == (10, 10)


# @pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
# def test_gaussian2d_batch(gaussians, show_plots):
#     """Inspect the test batch of 2D Gaussians."""
#     gaussians = gaussians * 1e15
#     for gaussian in gaussians:
#         gaussian = gaussian.squeeze().cpu().numpy()
#         plt.imshow(gaussian, interpolation='nearest')
#         if show_plots:
#             plt.show()
#     plt.close("all")

@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
@pytest.mark.parametrize(
    "mode,fixed_center",
    [
        ("centroid", None),
        ("image_center", None),
        ("fixed", (0.5, 0.5)),
    ],
)
def test_compute_centers(mode, fixed_center, gaussians, show_plots):
    """Test the compute_centers function."""
    X, Y = profile._make_grid(gaussians.shape[-1], gaussians.shape[-2])
    centers = profile.compute_centers(
        gaussians,
        X,
        Y,
        mode=mode,
        fixed_center=(0.5, 0.5),
    )
    assert centers.shape[0] == gaussians.shape[0]
    assert centers.shape[1] == 2
    if show_plots:
        for gaussian, center in zip(gaussians, centers):
            plt.imshow(gaussian.squeeze().cpu().numpy(), interpolation='nearest')
            plt.scatter(center[0].cpu().numpy(), center[1].cpu().numpy(), c='cyan')
            plt.show()


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_compute_centers_invalid_mode(gaussians):
    """Test compute_centers with an invalid mode."""
    X, Y = profile._make_grid(gaussians.shape[-1], gaussians.shape[-2])
    with pytest.raises(ValueError):
        profile.compute_centers(
            gaussians,
            X,
            Y,
            mode="invalid_mode",
        )


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_radial_histogram(gaussians, show_plots):
    """Test the radial histogram function."""
    gaussians = gaussians * 1e12
    hist, edges, counts = profile.radial_histogram(gaussians, nbins=256, average=True)
    for ih, h, c in zip(edges, hist, counts):
        h = h.cpu().numpy()
        ih = ih.cpu().numpy()
        if show_plots:
            plt.plot(0.5*(ih[:-1] + ih[1:]), h)
            plt.show()
    assert hist.shape[0] == gaussians.shape[0]
    assert hist.shape[1] == 256
    assert hist.shape == counts.shape
    assert edges.shape[1] == hist.shape[1] + 1


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_radial_histogram_invalid(gaussians):
    """Test radial_histogram with invalid inputs."""
    with pytest.raises(ValueError):
        profile.radial_histogram(gaussians, nbins=0)


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_cumulative_radial_histogram(gaussians, show_plots):
    """Test the cumulative radial histogram function."""
    gaussians = gaussians * 1e12
    hist, edges, counts = profile.cumulative_radial_histogram(gaussians, nbins=256)
    for ih, h, c in zip(edges, hist, counts):
        print(h[-1], torch.sum(c))
        h = h.cpu().numpy()
        ih = ih.cpu().numpy()
        if h.shape[0] > 0 and show_plots:
            plt.plot(0.5*(ih[:-1] + ih[1:]), h)
            plt.show()


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_radial_pdf(gaussians, show_plots):
    """Test the radial probability density function."""
    gaussians = gaussians * 1e12
    pdf, edges = profile.radial_pdf(gaussians, nbins=256)
    for ip, p in zip(edges, pdf):
        p = p.cpu().numpy()
        ip = ip.cpu().numpy()
        if p.shape[0] > 0 and show_plots:
            plt.plot(0.5*(ip[:-1] + ip[1:]), p)
            plt.show()


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_radial_pdf_invalid(gaussians):
    """Test the radial probability density function with invalid inputs."""
    with pytest.raises(ValueError):
        profile.radial_pdf(gaussians, nbins=0)


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_radial_cdf(gaussians, show_plots):
    """Test the radial cumulative density function."""
    gaussians = gaussians * 1e12
    cdf, edges = profile.radial_cdf(gaussians, nbins=256)
    for ic, c in zip(edges, cdf):
        print(c[-1])
        c = c.cpu().numpy()
        ic = ic.cpu().numpy()
        if c.shape[0] > 0 and show_plots:
            plt.plot(0.5*(ic[:-1] + ic[1:]), c)
            plt.show()


@pytest.mark.skipif(not TORCH_AVAILABLE, reason="torch not installed")
def test_RadialProfile(gaussians, show_plots):
    """Test the RadialProfile class."""
    gaussians = gaussians * 1e12
    rel_noise = torch.rand(*gaussians.shape)
    gaussians_noise = gaussians * rel_noise
    prof = profile.RadialProfile(nbins=100)
    prof.update(gaussians, gaussians_noise)
    prof.update(gaussians, gaussians_noise)
    assert prof.aggregate.shape[0] == gaussians.shape[0] * 2
    assert prof.aggregate.shape[1] == 100
    assert prof.aggregate.shape == prof.target_aggregrate.shape
    profile_summaries = prof.compute()
    assert profile_summaries[0].shape[0] == 100
    assert profile_summaries[0].shape == profile_summaries[1].shape
    if show_plots:
        for i in range(prof.aggregate.shape[0]):
            plt.plot(prof.edges[:-1], prof.aggregate[i].cpu().numpy(), label=f"Profile {i}")
        plt.show()



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""All utility functions can be imported from here."""

from chuchichaestli.data.cache import nbytes
from skais_mapper.utils.helper import current_time, compress_encode, extract_decode, alias_kw
from skais_mapper.utils.config import get_run_id, set_run_id
from skais_mapper.utils.primes import next_prime
from skais_mapper.utils.colors import SkaisColors, SkaisColorMaps


__all__ = [
    "nbytes",
    "current_time",
    "compress_encode",
    "extract_decode",
    "alias_kw",
    "get_run_id",
    "set_run_id",
    "next_prime",
    "SkaisColors",
    "SkaisColorMaps",
]



from skais_mapper._compat import OptionalDependencyNotAvailable

_MSG = (
    "This functionality requires the optional dependency 'torch'.\n"
    "Install the package with: pip install 'skais-mapper[nn]'"
)

def __getattr__(name: str):
    """Raise ImportError."""
    raise OptionalDependencyNotAvailable(_MSG)

def __dir__():
    """Make dir() somewhat informative without importing torch."""
    return []



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Testing the skais_mapper.cosmology module."""

import numpy as np
from skais_mapper.rotations import R

def test_init_and_identity_omega():
    """Test initialization and default omega."""
    # Default init (omega=None)
    r = R()
    assert r.omega is None
    arr = np.eye(3)
    result = r(arr)
    np.testing.assert_array_almost_equal(result, arr)
    assert r.omega is not None  # Should be set after first call


def test_call_with_given_omega():
    """"Test calling with a specific omega."""
    omega = np.eye(3)
    r = R(omega=omega)
    arr = np.eye(3)
    result = r(arr)
    np.testing.assert_array_almost_equal(result, arr)


def test_call_with_rotation():
    """Test calling with a rotation matrix."""
    theta = 90
    r = R.x(theta)
    arr = np.array([[0, 1, 0]])
    # Should rotate (0, 1, 0) to (0, 0, 1)
    rotated = r(arr)
    np.testing.assert_array_almost_equal(
        rotated, np.array([[0, 0, 1]]), decimal=6
    )


def test_mul_combines_rotations():
    """Test multiplication of two rotations."""
    r1 = R.x(90)
    r2 = R.y(90)
    r3 = r1 * r2
    # Should be type R and omega = r1.omega @ r2.omega
    assert isinstance(r3, R)
    np.testing.assert_array_almost_equal(r3.omega, r1.omega @ r2.omega)


def test_x_y_z_classmethods_and_static_equivalence():
    """Test class methods and static methods for x, y, z rotations."""
    theta = 45
    rx = R.x(theta)
    expected = R._x(theta)
    np.testing.assert_array_almost_equal(rx.omega, expected)
    ry = R.y(theta)
    expected = R._y(theta)
    np.testing.assert_array_almost_equal(ry.omega, expected)
    rz = R.z(theta)
    expected = R._z(theta)
    np.testing.assert_array_almost_equal(rz.omega, expected)


def test_rotation_static_degrees_vs_radians():
    """Test static methods for x, y, z rotations with degrees and radians."""
    theta_deg = 90
    theta_rad = np.pi / 2
    rx_deg = R._x(theta_deg, degrees=True)
    rx_rad = R._x(theta_rad, degrees=False)
    np.testing.assert_array_almost_equal(rx_deg, rx_rad)
    ry_deg = R._y(theta_deg, degrees=True)
    ry_rad = R._y(theta_rad, degrees=False)
    np.testing.assert_array_almost_equal(ry_deg, ry_rad)
    rz_deg = R._z(theta_deg, degrees=True)
    rz_rad = R._z(theta_rad, degrees=False)
    np.testing.assert_array_almost_equal(rz_deg, rz_rad)


def test_rotation_on_multiple_vectors():
    """Test rotation on multiple vectors."""
    r = R.x(90)
    arr = np.array([
        [0, 1, 0],
        [1, 0, 0],
        [0, 0, 1]
    ])
    expected = np.array([
        [0, 0, 1],
        [1, 0, 0],
        [0, -1, 0]
    ])
    np.testing.assert_array_almost_equal(r(arr), expected, decimal=6)
